#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <sys/time.h>
#include <unistd.h>
#include "udm_utils.h"
#include "udm_os.h"

int tz_offset=0;

char * UdmTrim(char *p, char *delim){
int len;
	len = strlen(p);
	while ((len > 0) && strchr(delim, p[len - 1] )) {
		p[len - 1] = '\0';
		len--;
	}
	while((*p)&&(strchr(delim,*p)))p++;
	return(p);
}

char * UdmRTrim(char* p, char *delim){
int len;
	len = strlen(p);
	while ((len > 0) && strchr(delim, p[len - 1] )) {
		p[len - 1] = '\0';
		len--;
	}
	return(p);
}


/* strtok_r clone */
char * UdmGetToken(char *s, const char *delim, char **last)
{
    char *spanp;
    int c, sc;
    char *tok;

    if (s == NULL && (s = *last) == NULL)
	return NULL;

cont:
    c = *s++;
    for (spanp = (char *)delim; (sc = *spanp++) != 0; )
    {
	if (c == sc)
	{
	    goto cont;
	}
    }

    if (c == 0)		/* no non-delimiter characters */
    {
	*last = NULL;
	return NULL;
    }
    tok = s - 1;

    for (;;)
    {
	c = *s++;
	spanp = (char *)delim;
	do
	{
	    if ((sc = *spanp++) == c)
	    {
		if (c == 0)
		{
		    s = NULL;
		}
		else
		{
		    char *w = s - 1;
		    *w = '\0';
		}
		*last = s;
		return tok;
	    }
	}
	while (sc != 0);
    }
}



char * UdmHtmlSpecialChars(char *str){
char *pos, *p;
int len;
	if(!str)return(NULL);
	len = strlen(str)+1;
	for(pos = str; *pos; pos++){
		switch (*pos){
			case '&': len+=4; break;
			case '"': len+=5; break;
			case '<':
			case '>': len+=3; break;
		}
	}
	p = (char*) malloc(len+1);*p = 0;
	for(pos=p; *str; str++) {
	switch (*str){
		case '&': strcat(p, "&amp;"); pos+=5; break;
		case '"': strcat(p, "&quot;");pos+=6; break;
		case '<': strcat(p, "&lt;");pos+=4; break;
		case '>': strcat(p, "&gt;");pos+=4; break;
		default:
			*pos = *str;
			pos++;*pos=0;
		}
	}
	return p;
}

char * UdmUnescapeCGIQuery(char *d,char *s){
int hi,lo=0;
char *dd;
	if((d==NULL)||(s==NULL))return(0);
	dd=d;
	while(*s){
		if(*s=='%'){
			if(strchr("0123456789",*(++s))) hi=*s-'0';
			else hi=*s-'A'+10;
			if(strchr("0123456789",*(++s))) lo=*s-'0';
			else lo=*s-'A'+10;
			*d=hi*16+lo;
		}else
		if(*s=='+'){
			*d=' ';
		}else{
			*d=*s;
		}
		s++; d++;
	}
	*d=0;return(dd);
}

char * UdmEscapeURL(char *d,char *s){
char *dd;
	if((d==NULL)||(s==NULL))return(0);
	dd=d;
	while(*s){
		if(strchr("%&<>+[](){}/?#'\"\\;,",*s)){
			sprintf(d,"%%%x",(int)*s);
			*(d+1)=toupper(*(d+1));
			*(d+2)=toupper(*(d+2));
			d+=2;
		}else
		if(*s==' '){
			*d='+';
		}else{
			*d=*s;
		}
		s++;d++;
	}
	*d=0;
	return(dd);
}


/* Oh,no! We have to remove parent level like /parent/../index.html from path
   Let's do it recursively! */
char * UdmRemove2Dot(char *path){
        char *ptr;
	char *tail;
    
        if(!(ptr=strstr(path,"../"))) return path;
	if(ptr==path) return path; /* How could it be? */
        tail=ptr+2;
	ptr--;
        *ptr=0;
	if(!(ptr=strrchr(path,'/'))) *path=0; else *ptr=0;
        path=strcat(path,tail);
	return UdmRemove2Dot(path);
}

/* This should convert Last-Modified time returned by webserver
 * to time_t (seconds since the Epoch). -kir
 */
time_t UdmHttpDate2Time_t(char * time_str){
/*
  HTTP applications have historically allowed three different formats
  for the representation of date/time stamps:
 
  Sun, 06 Nov 1994 08:49:37 GMT  ; RFC 822, updated by RFC 1123
  Sunday, 06-Nov-94 08:49:37 GMT ; RFC 850, obsoleted by RFC 1036
  Sun Nov  6 08:49:37 1994       ; ANSI C's asctime() format
			
  The first format is preferred as an Internet standard and represents
  a fixed-length subset of that defined by RFC 1123 [8] (an update to
  RFC 822 [9]). The second format is in common use, but is based on the
  obsolete RFC 850 [12] date format and lacks a four-digit year.
  HTTP/1.1 clients and servers that parse the date value MUST accept
  all three formats (for compatibility with HTTP/1.0), though they MUST
  only generate the RFC 1123 format for representing HTTP-date values
  in header fields. See section 19.3 for further information.
*/						
    
	struct tm tim;
    
        UDMMEMZERO(&tim, sizeof(tim));
	if (!strptime(time_str, "%a, %d %b %Y %T %Z", &tim))
		if (!strptime(time_str, "%a, %d-%b-%y %T %Z", &tim))
			if (!strptime(time_str, "%a %b %d %T %Y", &tim))
				return (time_t)-1; /* error */
	
	return mktime(&tim)-tz_offset;
}

time_t UdmFTPDate2Time_t(char *date){
        struct tm tm_s;

        if (strlen(date)!=20)
                return -1;
        UDMMEMZERO(&tm_s, sizeof(tm_s));
        strptime(date+6, "%y%m%d%H%M%S", &tm_s);
        return mktime(&tm_s)-tz_offset;
}

void UdmTime_t2HttpStr(time_t t, char *str){
    struct tm *time;
    if (t==0)
	*str='\0';
    else{
	/* FIXME: I suspect that gmtime IS NOT REENTRABLE!
	 */
	time=gmtime(&t);
	if (!strftime(str, UDM_MAXTIMESTRLEN, "%a, %d %b %Y  %H:%M:%S %Z", time))
	    *str='\0';
	/* end of BAD code */
    }
}

/* Find out tz_offset for the functions above
 */
int UdmInitTZ(void){
        time_t tt;
	struct tm *tms;

	tms=localtime(&tt);
	/* localtime sets the external variable timezone */
	tz_offset=(int)timezone;
	return 0;
#if 0

#ifdef HAVE_TM_GMTOFF
	time_t clock;
	struct tm *tim;
        
	clock=time(NULL);
	tim=localtime(&clock);
        tz_offset=tim->tm_gmtoff;
	return 0;
#else                                                                           
	struct tm *t;
	int days, hours, minutes;

	tt= time(NULL);
        gmt = *gmtime(&tt);
        t = localtime(&tt);
        days = t->tm_yday - gmt.tm_yday;
        hours = ((days < -1 ? 24 : 1 < days ? -24 : days * 24)
                + t->tm_hour - gmt.tm_hour);
        minutes=hours * 60 + t->tm_min - gmt.tm_min;
        tz_offset=minutes*60;
	return 0;
#endif /* HAVE_TM_GMTOFF */

#endif /* if 0 */

/* ONE MORE VARIANT - how many do we have?
    struct timezone tz;

    gettimeofday(NULL, &tz);
    tz_offset=tz.tz_minuteswest*60;
    return 0;
*/
}

/* Below was taken from log.c, slightly modified and heavily reorganized.
 *	--kir.
 */
#ifndef HAVE_STRNLEN
int udm_strnlen(const char *s, int max){
const char *e;
	e=s;
	while((*e)&&((e-s)<max))e++;
	return(e-s);
}
#else 
#define udm_strnlen strnlen
#endif /* ifndef HAVE_STRNLEN */

#ifndef HAVE_VSNPRINTF
/* we use this so that we can do without the ctype library */
#define is_digit(c)	((c) >= '0' && (c) <= '9')
#define ZEROPAD	1		/* pad with zero */
#define SIGN	2		/* unsigned/signed long */
#define PLUS	4		/* show plus */
#define SPACE	8		/* space if plus */
#define LEFT	16		/* left justified */
#define SPECIAL	32		/* 0x */
#define LARGE	64		/* use 'ABCDEF' instead of 'abcdef' */

#define do_div(n,base) ({ \
int __res; \
__res = ((unsigned long) n) % (unsigned) base; \
n = ((unsigned long) n) / (unsigned) base; \
__res; })

static int skip_atoi(const char **s){
	int i=0;

	while (is_digit(**s))
		i = i*10 + *((*s)++) - '0';
	return i;
}

static int number(long num, int base, int size, int precision ,int type)
{
	int buflen=0;
	char c,sign,tmp[66];
	const char *digits="0123456789abcdefghijklmnopqrstuvwxyz";
	int i;

	if (type & LARGE)
		digits = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ";
	if (type & LEFT)
		type &= ~ZEROPAD;
	if (base < 2 || base > 36)
		return 0;
	c = (type & ZEROPAD) ? '0' : ' ';
	sign = 0;
	if (type & SIGN) {
		if (num < 0) {
			sign = '-';
			num = -num;
			size--;
		} else if (type & PLUS) {
			sign = '+';
			size--;
		} else if (type & SPACE) {
			sign = ' ';
			size--;
		}
	}
	if (type & SPECIAL) {
		if (base == 16)
			size -= 2;
		else if (base == 8)
			size--;
	}
	i = 0;
	if (num == 0)
		tmp[i++]='0';
	else while (num != 0)
		tmp[i++] = digits[do_div(num,base)];
	if (i > precision)
		precision = i;
	size -= precision;
	if (!(type&(ZEROPAD+LEFT)))
		while(size-->0)
			buflen++;
	if (sign)
		buflen++;
	if (type & SPECIAL) {
		if (base==8)
			buflen++;
		else if (base==16) {
			buflen++;
			buflen++;
		}
	}
	if (!(type & LEFT))
		while (size-- > 0)
			buflen++;
	while (i < precision--)
		buflen++;
	while (i-- > 0)
		buflen++;
	while (size-- > 0)
		buflen++;
	return buflen;
}

int udm_vsnprintf_len(const char *fmt, va_list args)
{
	int buflen=0;
	int len;
	unsigned long num;
	int i, base;
	const char *s;

	int flags;		/* flags to number() */

	int field_width;	/* width of output field */
	int precision;		/* min. # of digits for integers; max
				   number of chars for from string */
	int qualifier;		/* 'h', 'l', or 'L' for integer fields */

	for (; *fmt ; ++fmt) {
		if (*fmt != '%') {
			buflen++;
			continue;
		}
			
		/* process flags */
		flags = 0;
		repeat:
			++fmt;		/* this also skips first '%' */
			switch (*fmt) {
				case '-': flags |= LEFT; goto repeat;
				case '+': flags |= PLUS; goto repeat;
				case ' ': flags |= SPACE; goto repeat;
				case '#': flags |= SPECIAL; goto repeat;
				case '0': flags |= ZEROPAD; goto repeat;
				}
		
		/* get field width */
		field_width = -1;
		if (is_digit(*fmt))
			field_width = skip_atoi(&fmt);
		else if (*fmt == '*') {
			++fmt;
			/* it's the next argument */
			field_width = va_arg(args, int);
			if (field_width < 0) {
				field_width = -field_width;
				flags |= LEFT;
			}
		}

		/* get the precision */
		precision = -1;
		if (*fmt == '.') {
			++fmt;	
			if (is_digit(*fmt))
				precision = skip_atoi(&fmt);
			else if (*fmt == '*') {
				++fmt;
				/* it's the next argument */
				precision = va_arg(args, int);
			}
			if (precision < 0)
				precision = 0;
		}

		/* get the conversion qualifier */
		qualifier = -1;
		if (*fmt == 'h' || *fmt == 'l' || *fmt == 'L') {
			qualifier = *fmt;
			++fmt;
		}

		/* default base */
		base = 10;

		switch (*fmt) {
		case 'c':
			if (!(flags & LEFT))
				while (--field_width > 0)
					buflen++;
			buflen+= sizeof((unsigned char) va_arg(args, int));
			while (--field_width > 0)
				buflen++;
			continue;

		case 's':
			s = va_arg(args, char *);
			if (!s)
				s = "<NULL>";

			len = udm_strnlen(s, precision);

			if (!(flags & LEFT))
				while (len < field_width--)
					buflen++;
			for (i = 0; i < len; ++i){
				buflen++;
				s++;
			}
			while (len < field_width--)
				buflen++;
			continue;

		case 'p':
			if (field_width == -1) {
				field_width = 2*sizeof(void *);
				flags |= ZEROPAD;
			}
			buflen += number(
				(unsigned long) va_arg(args, void *), 16,
				field_width, precision, flags);
			continue;


		case '%':
			buflen++;
			continue;

		/* integer number formats - set up the flags and "break" */
		case 'o':
			base = 8;
			break;

		case 'X':
			flags |= LARGE;
		case 'x':
			base = 16;
			break;

		case 'd':
		case 'i':
			flags |= SIGN;
		case 'u':
			break;

		default:
			buflen++;
			if (*fmt)
				buflen++;
			else
				--fmt;
			continue;
		}
		if (qualifier == 'l')
			num = va_arg(args, unsigned long);
		else if (qualifier == 'h') {
			num = (unsigned short) va_arg(args, int);
			if (flags & SIGN)
				num = (short) num;
		} else if (flags & SIGN)
			num = va_arg(args, int);
		else
			num = va_arg(args, unsigned int);
		buflen += number(num, base, field_width, precision, flags);
	}
	return buflen;
}

int udm_vsnprintf(char *str, size_t size, const char  *fmt,  va_list ap){
	if (udm_vsnprintf_len(fmt, ap)<size){
                return(vsprintf(str, fmt, ap));
        }else{
                strncpy(str, "Oops! Buffer overflow in udm_vsnprintf", size);
                return(-1);
        }
}
#else /* ifndef HAVE_VSNPRINTF */
#define udm_vsnprintf vsnprintf
#endif /* ifndef HAVE_VSNPRINTF */

/* This snprintf function has been written
 * by Murray Jensen <Murray.Jensen@cmst.csiro.au>
 * So if it works for you - praise him, 
 * and if it doesn't - blame him...  
 *   --kir.
 */
#ifndef HAVE_SNPRINTF
int udm_snprintf(char *buf, int len, const char *fmt, ...){
#ifndef HAVE_VSNPRINTF
       char *tmpbuf;
       int need;
#endif /* ifndef HAVE_VSNPRINTF */
       va_list ap;
       int ret;

       va_start(ap, fmt);

#ifdef HAVE_VSNPRINTF
       ret = vsnprintf(buf, len, fmt, ap);
#else
       buf[len-1] = '\0';

       need = udm_vsnprintf_len(fmt, ap);

       if ((tmpbuf = (char *)malloc(need + 1)) == NULL)
               strncpy(buf, "Oops! Out of memory in udm_snprintf", len-1);
       else {
               vsprintf(tmpbuf, fmt, ap);
               strncpy(buf, tmpbuf, len-1);
               free(tmpbuf);
       }

       ret = strlen(buf);
#endif /* HAVE_VSNPRINTF */

       va_end(ap);

       return ret;
}
#else
#define udm_snprintf snprintf
#endif /* HAVE_SNPRINTF */

